"""
OpenTC is text classification engine running as client server architecture. It listen on port 3333

"""
from opentc.core.setting import __version__


__author__ = "Cahya Wirawan <Cahya.Wirawan@gmail.com>"
